#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_commons.databases as commons_databases
import octobot_commons.tree as commons_tree

import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.storage.util as storage_util
import octobot_trading.personal_data.portfolios.history as portfolio_history


class PortfolioStorage(abstract_storage.AbstractStorage):
    IS_LIVE_CONSUMER = False
    IS_HISTORICAL = True
    PRICE_INIT_TIMEOUT = 30
    HISTORY_TABLE = commons_enums.RunDatabases.HISTORICAL_PORTFOLIO_VALUE.value

    @abstract_storage.AbstractStorage.hard_reset_and_retry_if_necessary
    async def store_history(self, reset=False):
        if not self.enabled:
            return
        portfolio_db = self.get_db()
        hist_portfolio_values_manager = self.exchange_manager.exchange_personal_data.\
            portfolio_manager.historical_portfolio_value_manager
        metadata = hist_portfolio_values_manager.get_metadata()
        # replace the whole table to ensure consistency
        history = hist_portfolio_values_manager.get_dict_historical_values()
        existing_history = await portfolio_db.all(self.HISTORY_TABLE)
        self._to_update_auth_data_ids_buffer.update(
            (
                history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY]
                for history_val in history
                if history_val not in existing_history
            )
        )
        await portfolio_db.upsert(commons_enums.RunDatabases.METADATA.value, metadata, None, uuid=1)
        await portfolio_db.replace_all(
            self.HISTORY_TABLE,
            history,
            cache=False
        )
        await self.trigger_debounced_flush()
        await self.trigger_debounced_update_auth_data(reset)

    async def _update_auth_data(self, reset):
        hist_portfolio_values_manager = self.exchange_manager.exchange_personal_data. \
            portfolio_manager.historical_portfolio_value_manager
        authenticator = authentication.Authenticator.instance()
        full_history = hist_portfolio_values_manager.get_dict_historical_values()
        history = [
            history_val
            for history_val in full_history
            if history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY] in self._to_update_auth_data_ids_buffer
        ]
        if full_history and authenticator.is_initialized():
            initializing_prices = hist_portfolio_values_manager.portfolio_manager.portfolio_value_holder.\
                value_converter.initializing_symbol_prices_pairs
            if initializing_prices:
                for symbol in initializing_prices:
                    await commons_tree.EventProvider.instance().wait_for_event(
                        self.exchange_manager.bot_id,
                        commons_tree.get_exchange_path(
                            self.exchange_manager.exchange_name,
                            commons_enums.InitializationEventExchangeTopics.PRICE.value,
                            symbol=symbol,
                        ),
                        self.PRICE_INIT_TIMEOUT
                    )
            await authenticator.update_portfolio(
                full_history[-1][portfolio_history.HistoricalAssetValue.VALUES_KEY],
                full_history[0][portfolio_history.HistoricalAssetValue.VALUES_KEY],
                float(hist_portfolio_values_manager.portfolio_manager.portfolio_profitability.profitability_percent),
                hist_portfolio_values_manager.portfolio_manager.reference_market,
                hist_portfolio_values_manager.ending_portfolio,
                {
                    history_val[portfolio_history.HistoricalAssetValue.TIMESTAMP_KEY]:
                        history_val[portfolio_history.HistoricalAssetValue.VALUES_KEY]
                    for history_val in history
                },
                hist_portfolio_values_manager.portfolio_manager.portfolio_value_holder.current_crypto_currencies_values,
                reset
            )
            self._to_update_auth_data_ids_buffer.clear()

    def get_db(self):
        return self._get_db()

    def _get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_historical_portfolio_value_db(
            self.exchange_manager.bot_id,
            storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager),
            self.exchange_manager.exchange_name,
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.signals as signals
import octobot_trading.signals.util as signal_util
import octobot_trading.enums as trading_enums
import octobot_trading.errors as trading_errors
import octobot_trading.constants as trading_constants


class TradingSignalBundleBuilder(signals.SignalBundleBuilder):
    def __init__(self, identifier: str, strategy: str):
        super().__init__(identifier)
        self.strategy = strategy

    def build(self) -> signals.SignalBundle:
        """
        Link bundled an grouped orders and create a signal_bundle.SignalBundle from registered signals
        """
        self._pack_referenced_orders_together()
        return super().build()

    def sort_signals(self):
        # https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts
        # move cancelled order signals at the beginning of the list, leave others as is while keeping initial order
        def _sort_key(signal):
            return 0 if signal.content[trading_enums.TradingSignalCommonsAttrs.ACTION.value] == \
                trading_enums.TradingSignalOrdersActions.CANCEL.value else 1

        self.signals = sorted(self.signals, key=_sort_key)
        return self

    def add_created_order(self, order, exchange_manager, target_amount=None, target_position=None):
        if target_amount is None and target_position is None:
            raise trading_errors.InvalidArgumentError("target_amount or target_position has to be provided")
        if not self._update_pending_orders(order, exchange_manager,
                                           trading_enums.TradingSignalOrdersActions.CREATE,
                                           target_amount=target_amount,
                                           target_position=target_position):
            self.register_signal(
                trading_enums.TradingSignalTopics.ORDERS.value,
                signal_util.create_order_signal_content(
                    order,
                    trading_enums.TradingSignalOrdersActions.CREATE,
                    self.strategy,
                    exchange_manager,
                    target_amount=target_amount,
                    target_position=target_position
                )
            )

    def add_order_to_group(self, order, exchange_manager):
        if not self._update_pending_orders(order, exchange_manager,
                                           trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP):
            self.register_signal(
                trading_enums.TradingSignalTopics.ORDERS.value,
                signal_util.create_order_signal_content(
                    order,
                    trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP,
                    self.strategy,
                    exchange_manager,
                )
            )

    def add_edited_order(
            self, order, exchange_manager,
            updated_target_amount=None,
            updated_target_position=None,
            updated_limit_price=trading_constants.ZERO,
            updated_stop_price=trading_constants.ZERO,
            updated_current_price=trading_constants.ZERO
    ):
        if updated_target_amount is updated_target_position is None and \
           updated_limit_price is updated_stop_price is updated_current_price is trading_constants.ZERO:
            raise trading_errors.InvalidArgumentError("an updated argument has to be provided")
        if not self._update_pending_orders(
                order,
                exchange_manager,
                trading_enums.TradingSignalOrdersActions.EDIT,
                updated_target_amount=updated_target_amount,
                updated_target_position=updated_target_position,
                updated_limit_price=updated_limit_price,
                updated_stop_price=updated_stop_price,
                updated_current_price=updated_current_price
        ):
            self.register_signal(
                trading_enums.TradingSignalTopics.ORDERS.value,
                signal_util.create_order_signal_content(
                    order,
                    trading_enums.TradingSignalOrdersActions.EDIT,
                    self.strategy,
                    exchange_manager,
                    updated_target_amount=updated_target_amount,
                    updated_target_position=updated_target_position,
                    updated_limit_price=updated_limit_price,
                    updated_stop_price=updated_stop_price,
                    updated_current_price=updated_current_price
                )
            )

    def add_cancelled_order(self, order, exchange_manager):
        if not self._update_pending_orders(order, exchange_manager, trading_enums.TradingSignalOrdersActions.CANCEL):
            self.register_signal(
                trading_enums.TradingSignalTopics.ORDERS.value,
                signal_util.create_order_signal_content(
                    order,
                    trading_enums.TradingSignalOrdersActions.CANCEL,
                    self.strategy,
                    exchange_manager,
                )
            )

    def _update_pending_orders(
        self, order, exchange_manager,
        action,
        target_amount=None,
        target_position=None,
        updated_target_amount=None,
        updated_target_position=None,
        updated_limit_price=trading_constants.ZERO,
        updated_stop_price=trading_constants.ZERO,
        updated_current_price=trading_constants.ZERO,
    ):
        try:
            index, order_description = self._get_order_description_from_local_orders(order.order_id)
            if action is trading_enums.TradingSignalOrdersActions.CREATE:
                # replace order
                self.signals[index] = self.create_signal(
                    trading_enums.TradingSignalTopics.ORDERS.value,
                    signal_util.create_order_signal_content(
                        order,
                        trading_enums.TradingSignalOrdersActions.CREATE,
                        self.strategy,
                        exchange_manager,
                        target_amount=target_amount,
                        target_position=target_position
                    )
                )
            if action is trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP:
                # update order group
                order_description[trading_enums.TradingSignalOrdersAttrs.GROUP_ID.value] = order.order_group.name
                order_description[trading_enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] = \
                    order.order_group.__class__.__name__
            elif action is trading_enums.TradingSignalOrdersActions.EDIT:
                # avoid editing order that are not yet created
                order_description[trading_enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] = \
                    updated_target_amount
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] = \
                    updated_target_amount
                order_description[trading_enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] = \
                    updated_target_position
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] = \
                    updated_target_position
                order_description[trading_enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value] = \
                    float(updated_limit_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] = \
                    float(updated_limit_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.STOP_PRICE.value] = \
                    float(updated_stop_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] = \
                    float(updated_stop_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value] = \
                    float(updated_current_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] = \
                    float(updated_current_price)
            elif action is trading_enums.TradingSignalOrdersActions.CANCEL:
                if order_description[trading_enums.TradingSignalCommonsAttrs.ACTION.value] == \
                   trading_enums.TradingSignalOrdersActions.CREATE.value:
                    # avoid creating order that are to be cancelled
                    self.signals.pop(index)
                else:
                    # now cancel order (no need to perform previous actions as it will get cancelled anyway
                    order_description[trading_enums.TradingSignalCommonsAttrs.ACTION.value] = \
                        trading_enums.TradingSignalOrdersActions.CANCEL.value
            return True
        except trading_errors.OrderDescriptionNotFoundError:
            pass
        return False

    def _pack_referenced_orders_together(self):
        filtered_signals = []
        order_description_by_seen_group_ids = {}
        for signal in self.signals:
            include_signal = True
            try:
                # bundled orders must be sent at the same time as the original order, add them both to the same signal
                if add_to_order_id := signal.content[trading_enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value]:
                    _, order_description = self._get_order_description_from_local_orders(add_to_order_id)
                    order_description[trading_enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value]\
                        .append(signal.content)
                    include_signal = False
                # chained orders must be sent at the same time as the original order, add them both to the same signal
                elif add_to_order_id := signal.content[trading_enums.TradingSignalOrdersAttrs.CHAINED_TO.value]:
                    _, order_description = self._get_order_description_from_local_orders(add_to_order_id)
                    order_description[trading_enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value]\
                        .append(signal.content)
                    include_signal = False
            except trading_errors.OrderDescriptionNotFoundError:
                # Failed to find the triggering order's associated signal that this order signal should be associated to
                # in ADDITIONAL_ORDERS.
                # This can happen when the triggering order has already been sent in a different signal
                self.logger.debug(f"Skip signal packing for {signal}: triggering order not found in associated signals")
            # grouped orders must be created before the group is enabled, add them both to the same signal
            # groups. They might also need to be updated together
            if (add_to_group_ids := signal.content[trading_enums.TradingSignalOrdersAttrs.GROUP_ID.value]) \
               and include_signal:
                if add_to_group_ids in order_description_by_seen_group_ids:
                    order_description_by_seen_group_ids[add_to_group_ids][
                        trading_enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value].append(signal.content)
                    include_signal = False
                else:
                    order_description_by_seen_group_ids[add_to_group_ids] = signal.content
            if include_signal:
                filtered_signals.append(signal)
        self.signals = filtered_signals

    def _get_order_description_from_local_orders(self, order_id):
        for index, signal in enumerate(self.signals):
            if signal.content[trading_enums.TradingSignalOrdersAttrs.ORDER_ID.value] == order_id:
                return index, signal.content
        raise trading_errors.OrderDescriptionNotFoundError(
            f"order not found (order_id: {order_id})"
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import numpy as np

import octobot_commons.enums as enums
import octobot_commons.constants as commons_constants

import octobot_trading.exchange_data.ohlcv.candles_manager as candles_manager


class PreloadedCandlesManager(candles_manager.CandlesManager):

    def get_preloaded_symbol_candles_count(self):
        return len(self.time_candles)

    def get_preloaded_symbol_close_candles(self):
        return self.close_candles

    def get_preloaded_symbol_open_candles(self):
        return self.open_candles

    def get_preloaded_symbol_high_candles(self):
        return self.high_candles

    def get_preloaded_symbol_low_candles(self):
        return self.low_candles

    def get_preloaded_symbol_time_candles(self):
        return self.time_candles

    def get_preloaded_symbol_volume_candles(self):
        return self.volume_candles

    def _set_all_candles(self, new_candles_data):
        self.close_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_CLOSE.value)
        self.open_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_OPEN.value)
        self.high_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_HIGH.value)
        self.low_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_LOW.value)
        self.time_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_TIME.value)
        self.volume_candles = self._get_candle_values_array(new_candles_data, enums.PriceIndexes.IND_PRICE_VOL.value)

    def _get_candle_values_array(self, candles, key):
        return np.array([candle[key] for candle in candles], dtype=np.float64)

    def _get_candle_index(self, candle):
        # Uses the given candle to find the index on the associated candle in preloaded candles.
        # The goal of this method is to quickly identify where the limit between past and future candles
        # should be when handling preloaded candles.

        # return actual index + 1 as it is used as a select length
        select_index = 0 if self.time_candles_index == 0 else self.time_candles_index - 1
        for delta_index, time_value in enumerate(self.time_candles[select_index:]):
            if time_value == candle[enums.PriceIndexes.IND_PRICE_TIME.value]:
                return self.time_candles_index + delta_index
        # candle in past candles
        for index, time_value in enumerate(self.time_candles[:select_index]):
            if time_value == candle[enums.PriceIndexes.IND_PRICE_TIME.value]:
                return index
        return commons_constants.DEFAULT_IGNORED_VALUE

    def add_old_and_new_candles(self, candles_data):
        # candles are already loaded, just set indexes to the new candle
        current_index = self._get_candle_index(candles_data[-1])
        if current_index == commons_constants.DEFAULT_IGNORED_VALUE:
            self.logger.error(
                f"Can't find candle at time: {candles_data[-1][enums.PriceIndexes.IND_PRICE_TIME.value]}"
            )
            return
        self.close_candles_index = current_index
        self.open_candles_index = current_index
        self.high_candles_index = current_index
        self.low_candles_index = current_index
        self.time_candles_index = current_index
        self.volume_candles_index = current_index

    def _extract_limited_data(self, data, limit=-1, max_limit=-1):
        if limit == -1:
            if max_limit == -1:
                return data
            return np.array(data[:max_limit], dtype=np.float64)

        if max_limit == -1:
            return np.array(data[-min(limit, len(data)):], dtype=np.float64)
        else:
            return np.array(data[max(0, max_limit - limit): max_limit], dtype=np.float64)

    def add_new_candle(self, new_candle_data):
        self.logger.error("add_new_candle should not be called")

    def _reset_candles(self):
        self.candles_initialized = False

        self.close_candles_index = 0
        self.open_candles_index = 0
        self.high_candles_index = 0
        self.low_candles_index = 0
        self.time_candles_index = 0
        self.volume_candles_index = 0

        self.close_candles = np.ndarray((0,))
        self.open_candles = np.ndarray((0,))
        self.high_candles = np.ndarray((0,))
        self.low_candles = np.ndarray((0,))
        self.time_candles = np.ndarray((0,))
        self.volume_candles = np.ndarray((0,))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.exchanges.adapters as adapters
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums
import octobot_trading.exchanges.connectors.ccxt.constants as ccxt_constants
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants
import octobot_trading.enums as enums
from octobot_trading.enums import ExchangeConstantsOrderColumns as ecoc
import octobot_commons.enums as common_enums
import octobot_commons.constants as common_constants


class CCXTAdapter(adapters.AbstractAdapter):
    def fix_order(self, raw, **kwargs):
        fixed = super().fix_order(raw, **kwargs)
        try:
            exchange_timestamp = fixed[ecoc.TIMESTAMP.value]
            fixed[ecoc.TIMESTAMP.value] = \
                self.get_uniformized_timestamp(exchange_timestamp)
        except KeyError as e:
            self.logger.error(f"Fail to cleanup order dict ({e})")
        self._register_exchange_fees(fixed)
        return fixed

    def parse_order(self, fixed, **kwargs):
        # CCXT standard order parsing logic
        return fixed

    def _register_exchange_fees(self, order_or_trade):
        try:
            fees = order_or_trade[enums.ExchangeConstantsOrderColumns.FEE.value]
            fees[enums.FeePropertyColumns.EXCHANGE_ORIGINAL_COST.value] = fees[enums.FeePropertyColumns.COST.value]
            fees[enums.FeePropertyColumns.IS_FROM_EXCHANGE.value] = True
        except (KeyError, TypeError):
            pass

    def _fix_ohlcv_prices(self, ohlcv):
        for index, value in enumerate(ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value + 1:]):
            ohlcv[index + 1] = float(value)

    def fix_ohlcv(self, raw, **kwargs):
        fixed = super().fix_ohlcv(raw, **kwargs)
        # ensure open time is not the current time but the actual candle open time
        # time_frame kwarg has to be passed to parse candle time
        candles_s = 1
        if "time_frame" in kwargs:
            candles_s = common_enums.TimeFramesMinutes[common_enums.TimeFrames(kwargs["time_frame"])] * \
                        common_constants.MINUTE_TO_SECONDS
        for index, ohlcv in enumerate(fixed):
            try:
                int_val = int(self.get_uniformized_timestamp(ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value]))
                ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value] = int_val - (int_val % candles_s)
                self._fix_ohlcv_prices(ohlcv)
            except TypeError:
                # the last candle might not be properly set
                if self.connector.exchange_manager.exchange.DUMP_INCOMPLETE_LAST_CANDLE and index == len(fixed) - 1:
                    return fixed[:-1]
                raise
            except KeyError as e:
                self.logger.error(f"Fail to fix ohlcv ({e})")
        return fixed

    def parse_ohlcv(self, fixed, **kwargs):
        # CCXT standard ohlcv parsing logic
        return fixed

    def fix_kline(self, raw, **kwargs):
        fixed = super().fix_kline(raw, **kwargs)
        for index, kline in enumerate(fixed):
            try:
                kline[common_enums.PriceIndexes.IND_PRICE_TIME.value] = \
                    int(self.get_uniformized_timestamp(kline[common_enums.PriceIndexes.IND_PRICE_TIME.value]))
                self._fix_ohlcv_prices(kline)
            except TypeError:
                # the last candle might not be properly set
                if self.connector.exchange_manager.exchange.DUMP_INCOMPLETE_LAST_CANDLE and index == len(fixed) - 1:
                    return fixed[:-1]
                raise
            except KeyError as e:
                self.logger.error(f"Fail to fix kline ({e})")
        return fixed

    def parse_kline(self, fixed, **kwargs):
        # CCXT standard kline parsing logic
        return fixed

    def fix_ticker(self, raw, **kwargs):
        fixed = super().fix_ticker(raw, **kwargs)
        # CCXT standard ticker fixing logic
        if timestamp := fixed.get(enums.ExchangeConstantsTickersColumns.TIMESTAMP.value):
            fixed[enums.ExchangeConstantsTickersColumns.TIMESTAMP.value] = \
                int(self.get_uniformized_timestamp(timestamp))
        return fixed

    def parse_ticker(self, fixed, **kwargs):
        # CCXT standard ticker parsing logic
        return fixed

    def create_ticker_from_kline(self, kline, symbol, **kwargs):
        return {
            enums.ExchangeConstantsTickersColumns.SYMBOL.value: symbol,
            enums.ExchangeConstantsTickersColumns.TIMESTAMP.value: kline[common_enums.PriceIndexes.IND_PRICE_TIME.value],
            enums.ExchangeConstantsTickersColumns.OPEN.value: kline[common_enums.PriceIndexes.IND_PRICE_OPEN.value],
            enums.ExchangeConstantsTickersColumns.HIGH.value: kline[common_enums.PriceIndexes.IND_PRICE_HIGH.value],
            enums.ExchangeConstantsTickersColumns.LOW.value: kline[common_enums.PriceIndexes.IND_PRICE_LOW.value],
            enums.ExchangeConstantsTickersColumns.CLOSE.value: kline[common_enums.PriceIndexes.IND_PRICE_CLOSE.value],
            enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value: kline[common_enums.PriceIndexes.IND_PRICE_VOL.value],
            enums.ExchangeConstantsTickersColumns.LAST.value: kline[common_enums.PriceIndexes.IND_PRICE_CLOSE.value],
            enums.ExchangeConstantsTickersColumns.BID.value: None,
            enums.ExchangeConstantsTickersColumns.BID_VOLUME.value: None,
            enums.ExchangeConstantsTickersColumns.ASK.value: None,
            enums.ExchangeConstantsTickersColumns.ASK_VOLUME.value: None,
            enums.ExchangeConstantsTickersColumns.PREVIOUS_CLOSE.value: None,
        }

    def fix_balance(self, raw, **kwargs):
        fixed = super().fix_balance(raw, **kwargs)
        # remove not currency specific keys
        return fixed

    def parse_balance(self, fixed, **kwargs):
        fixed.pop(constants.CONFIG_PORTFOLIO_FREE, None)
        fixed.pop(constants.CONFIG_PORTFOLIO_USED, None)
        fixed.pop(constants.CONFIG_PORTFOLIO_TOTAL, None)
        fixed.pop(ccxt_constants.CCXT_INFO, None)
        fixed.pop(ccxt_enums.ExchangeConstantsCCXTColumns.DATETIME.value, None)
        fixed.pop(ccxt_enums.ExchangeConstantsCCXTColumns.TIMESTAMP.value, None)
        return personal_data.parse_decimal_portfolio(fixed)

    def fix_order_book(self, raw, **kwargs):
        fixed = super().fix_order_book(raw, **kwargs)
        # CCXT standard order_book fixing logic
        return fixed

    def parse_order_book(self, fixed, **kwargs):
        # CCXT standard order_book parsing logic
        return fixed

    def fix_public_recent_trades(self, raw, **kwargs):
        fixed = super().fix_public_recent_trades(raw, **kwargs)
        # CCXT standard public_recent_trades fixing logic
        for recent_trade in fixed:
            try:
                recent_trade[ecoc.TIMESTAMP.value] = \
                    self.get_uniformized_timestamp(recent_trade[ecoc.TIMESTAMP.value])
            except KeyError as e:
                self.logger.error(f"Fail to clean recent_trade dict ({e})")
        return fixed

    def parse_public_recent_trades(self, fixed, **kwargs):
        # CCXT standard public_recent_trades parsing logic
        for recent_trade in fixed:
            recent_trade.pop(ecoc.INFO.value, None)
            recent_trade.pop(ecoc.DATETIME.value, None)
            recent_trade.pop(ecoc.EXCHANGE_ID.value, None)
            recent_trade.pop(ecoc.ID.value, None)
            recent_trade.pop(ecoc.ORDER.value, None)
            recent_trade.pop(ecoc.FEE.value, None)
            recent_trade.pop(ecoc.TYPE.value, None)
            recent_trade.pop(ecoc.TAKER_OR_MAKER.value, None)
        return fixed

    def fix_trades(self, raw, **kwargs):
        fixed = super().fix_trades(raw, **kwargs)
        # CCXT standard trades fixing logic
        for trade in fixed:
            try:
                trade[ecoc.TIMESTAMP.value] = \
                    self.get_uniformized_timestamp(trade[ecoc.TIMESTAMP.value])
                trade[ecoc.EXCHANGE_ID.value] = trade.get(ecoc.ORDER.value)
                if trade[enums.ExchangeConstantsOrderColumns.TYPE.value] is None:
                    trade[enums.ExchangeConstantsOrderColumns.TYPE.value] = enums.TradeOrderType.MARKET.value \
                        if trade[ccxt_enums.ExchangeOrderCCXTColumns.TAKER_OR_MAKER.value] \
                        == enums.ExchangeConstantsMarketPropertyColumns.TAKER.value \
                        else enums.TradeOrderType.LIMIT.value
            except KeyError as e:
                self.logger.error(f"Fail to clean trade dict ({e})")
            self._register_exchange_fees(trade)
        return fixed

    def parse_trades(self, fixed, **kwargs):
        # CCXT standard trades parsing logic
        return fixed

    def fix_position(self, raw, **kwargs):
        fixed = super().fix_position(raw, **kwargs)
        # CCXT standard position fixing logic
        return fixed

    def parse_position(self, fixed, force_empty=False, **kwargs):
        # CCXT standard position parsing logic
        # if mode is enums.PositionMode.ONE_WAY:
        original_side = fixed.get(ccxt_enums.ExchangePositionCCXTColumns.SIDE.value)
        position_side = enums.PositionSide.BOTH
        # todo when handling cross positions
        # side = fixed.get(ccxt_enums.ExchangePositionCCXTColumns.SIDE.value, enums.PositionSide.UNKNOWN.value)
        # position_side = enums.PositionSide.LONG \
        #     if side == enums.PositionSide.LONG.value else enums.PositionSide.
        symbol = fixed.get(ccxt_enums.ExchangePositionCCXTColumns.SYMBOL.value)
        contract_size = decimal.Decimal(str(fixed.get(ccxt_enums.ExchangePositionCCXTColumns.CONTRACT_SIZE.value, 0)))
        contracts = constants.ZERO if force_empty \
            else decimal.Decimal(str(fixed.get(ccxt_enums.ExchangePositionCCXTColumns.CONTRACTS.value, 0)))
        is_empty = contracts == constants.ZERO
        liquidation_price = fixed.get(ccxt_enums.ExchangePositionCCXTColumns.LIQUIDATION_PRICE.value, 0)
        if margin_type := fixed.get(ccxt_enums.ExchangePositionCCXTColumns.MARGIN_TYPE.value, None):
            margin_type = enums.MarginType(margin_type)
        if force_empty or liquidation_price is None:
            liquidation_price = constants.NaN
        else:
            liquidation_price = decimal.Decimal(str(liquidation_price))
        try:
            fixed.update({
                enums.ExchangeConstantsPositionColumns.SYMBOL.value: symbol,
                enums.ExchangeConstantsPositionColumns.TIMESTAMP.value:
                    fixed.get(ccxt_enums.ExchangePositionCCXTColumns.TIMESTAMP.value,
                              self.connector.get_exchange_current_time()),
                enums.ExchangeConstantsPositionColumns.SIDE.value: position_side,
                enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value: margin_type,
                enums.ExchangeConstantsPositionColumns.SIZE.value:
                    contract_size * contracts if original_side == enums.PositionSide.LONG.value
                    else -contract_size * contracts,
                enums.ExchangeConstantsPositionColumns.CONTRACT_TYPE.value:
                    self.connector.exchange_manager.exchange.get_contract_type(symbol),
                enums.ExchangeConstantsPositionColumns.LEVERAGE.value:
                    self.safe_decimal(fixed, ccxt_enums.ExchangePositionCCXTColumns.LEVERAGE.value,
                                      constants.DEFAULT_SYMBOL_LEVERAGE),
                enums.ExchangeConstantsPositionColumns.POSITION_MODE.value: None if is_empty else
                enums.PositionMode.HEDGE if fixed.get(ccxt_enums.ExchangePositionCCXTColumns.HEDGED.value, True)
                else enums.PositionMode.ONE_WAY,
                # next values are always 0 when the position empty (0 contracts)
                enums.ExchangeConstantsPositionColumns.COLLATERAL.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.COLLATERAL.value, 0)}"),
                enums.ExchangeConstantsPositionColumns.NOTIONAL.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.NOTIONAL.value, 0)}"),
                enums.ExchangeConstantsPositionColumns.INITIAL_MARGIN.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.INITIAL_MARGIN.value, 0)}"),
                enums.ExchangeConstantsPositionColumns.UNREALIZED_PNL.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.UNREALISED_PNL.value, 0)}"),
                enums.ExchangeConstantsPositionColumns.REALISED_PNL.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.REALISED_PNL.value, 0) or 0}"),
                enums.ExchangeConstantsPositionColumns.LIQUIDATION_PRICE.value: liquidation_price,
                enums.ExchangeConstantsPositionColumns.MARK_PRICE.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.MARK_PRICE.value, 0)}"),
                enums.ExchangeConstantsPositionColumns.ENTRY_PRICE.value: constants.ZERO if is_empty else
                decimal.Decimal(
                    f"{fixed.get(ccxt_enums.ExchangePositionCCXTColumns.ENTRY_PRICE.value, 0)}"),
            })
        except KeyError as e:
            self.logger.error(f"Fail to parse position dict ({e})")
        return fixed

    def fix_funding_rate(self, raw, **kwargs):
        fixed = super().fix_funding_rate(raw, **kwargs)
        # CCXT standard funding_rate fixing logic
        return fixed

    def parse_funding_rate(self, fixed, from_ticker=False, **kwargs):
        # CCXT standard funding_rate parsing logic
        fixed.update({
            enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value:
                self.safe_decimal(
                    fixed, ccxt_enums.ExchangeFundingCCXTColumns.PREVIOUS_FUNDING_RATE.value, constants.NaN
                ),
            enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value:
                self.get_uniformized_timestamp(
                    fixed.get(ccxt_enums.ExchangeFundingCCXTColumns.PREVIOUS_FUNDING_TIMESTAMP.value, 0)
                    or constants.ZERO
                ),
            enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value:
                self.safe_decimal(
                    fixed, ccxt_enums.ExchangeFundingCCXTColumns.FUNDING_RATE.value, constants.NaN
                ),
            enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value:
                self.get_uniformized_timestamp(
                    fixed.get(ccxt_enums.ExchangeFundingCCXTColumns.FUNDING_TIMESTAMP.value, 0) or constants.ZERO
                ),
        })
        return fixed

    def fix_leverage(self, raw, **kwargs):
        fixed = super().fix_leverage(raw, **kwargs)
        # CCXT standard leverage fixing logic
        return fixed

    def parse_leverage(self, fixed, **kwargs):
        # WARNING no CCXT standard leverage parsing logic
        # HAS TO BE IMPLEMENTED IN EACH EXCHANGE IMPLEMENTATION
        return {
            enums.ExchangeConstantsLeveragePropertyColumns.RAW.value: fixed,
        }

    def fix_funding_rate_history(self, raw, **kwargs):
        fixed = super().fix_funding_rate_history(raw, **kwargs)
        # CCXT standard funding_rate_history fixing logic
        return fixed

    def parse_funding_rate_history(self, fixed, **kwargs):
        # CCXT standard funding_rate_history parsing logic
        return fixed

    def fix_leverage_tiers(self, raw, **kwargs):
        fixed = super().fix_leverage_tiers(raw, **kwargs)
        # CCXT standard leverage_tiers fixing logic
        return fixed

    def parse_leverage_tiers(self, fixed, **kwargs):
        # CCXT standard leverage_tiers parsing logic
        
        # {
        #   "BTC/USDT": [
        #       {
        #           ExchangeConstantsLeverageTiersColumns.TIER.value: 1.0, 
        #           ExchangeConstantsLeverageTiersColumns.CURRENCY.value: "USDT", 
        #           ExchangeConstantsLeverageTiersColumns.MIN_NOTIONAL.value: 0.0, 
        #           ExchangeConstantsLeverageTiersColumns.MAX_NOTIONAL.value: 5000.0,
        #           ExchangeConstantsLeverageTiersColumns.MAINTENANCE_MARGIN_RATE.value: 0.01, 
        #           ExchangeConstantsLeverageTiersColumns.MAX_LEVERAGE.value: 50.0, 
        #           ExchangeConstantsLeverageTiersColumns.INFO.value: {
        #               'bracket': '1', 'initialLeverage': '50', 'notionalCap': '5000',
        #               'notionalFloor': '0', 'maintMarginRatio': '0.01', 'cum': '0.0'
        #           }
        #       }
        #   ]
        # }
        for symbol, tiers in fixed.items():
            for tier_index, tier in enumerate(tiers):
                fixed[symbol][tier_index]= {
                    enums.ExchangeConstantsLeverageTiersColumns.TIER.value:
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.TIER.value),
                    enums.ExchangeConstantsLeverageTiersColumns.CURRENCY.value: 
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.CURRENCY.value),
                    enums.ExchangeConstantsLeverageTiersColumns.MIN_NOTIONAL.value:
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.MIN_NOTIONAL.value),
                    enums.ExchangeConstantsLeverageTiersColumns.MAX_NOTIONAL.value:
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.MAX_NOTIONAL.value),
                    enums.ExchangeConstantsLeverageTiersColumns.MAINTENANCE_MARGIN_RATE.value:
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.MAINTENANCE_MARGIN_RATE.value),
                    enums.ExchangeConstantsLeverageTiersColumns.MAX_LEVERAGE.value:
                        tier.get(ccxt_enums.ExchangeLeverageTiersCCXTColumns.MAX_LEVERAGE.value),
                }
        return fixed

    def fix_mark_price(self, raw, **kwargs):
        fixed = super().fix_mark_price(raw, **kwargs)
        # CCXT standard mark_price fixing logic
        return fixed

    def parse_mark_price(self, fixed, **kwargs):
        # CCXT standard mark_price parsing logic
        return fixed

    def fix_market_status(self, raw, **kwargs):
        fixed = super().fix_market_status(raw, **kwargs)
        # CCXT standard market_status fixing logic
        return fixed

    def parse_market_status(self, fixed, **kwargs):
        # CCXT standard market_status parsing logic
        return fixed



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import numpy as np

import octobot_commons.data_util as data_util
import octobot_commons.enums as enums
import octobot_commons.logging as logging

import octobot_trading.util as util


class CandlesManager(util.Initializable):
    MAX_CANDLES_COUNT = 1000

    def __init__(self, max_candles_count=None):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)

        self.candles_initialized = False
        self.max_candles_count = max_candles_count \
            if max_candles_count and max_candles_count > self.__class__.MAX_CANDLES_COUNT \
            else self.__class__.MAX_CANDLES_COUNT

        self.close_candles_index = 0
        self.open_candles_index = 0
        self.high_candles_index = 0
        self.low_candles_index = 0
        self.time_candles_index = 0
        self.volume_candles_index = 0

        self.close_candles = None
        self.open_candles = None
        self.high_candles = None
        self.low_candles = None
        self.time_candles = None
        self.volume_candles = None

        self.reached_max = False
        self._reset_candles()

    async def initialize_impl(self):
        self._reset_candles()

    def _reset_candles(self):
        self.candles_initialized = False
        self.reached_max = False

        self.close_candles_index = 0
        self.open_candles_index = 0
        self.high_candles_index = 0
        self.low_candles_index = 0
        self.time_candles_index = 0
        self.volume_candles_index = 0

        self.close_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)
        self.open_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)
        self.high_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)
        self.low_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)
        self.time_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)
        self.volume_candles = np.full(self.max_candles_count, fill_value=np.nan, dtype=np.float64)

    # getters
    def get_symbol_candles_count(self):
        return self.time_candles_index

    def get_symbol_close_candles(self, limit=-1):
        return self._extract_limited_data(self.close_candles, limit, max_limit=self.close_candles_index)

    def get_symbol_open_candles(self, limit=-1):
        return self._extract_limited_data(self.open_candles, limit, max_limit=self.open_candles_index)

    def get_symbol_high_candles(self, limit=-1):
        return self._extract_limited_data(self.high_candles, limit, max_limit=self.high_candles_index)

    def get_symbol_low_candles(self, limit=-1):
        return self._extract_limited_data(self.low_candles, limit, max_limit=self.low_candles_index)

    def get_symbol_time_candles(self, limit=-1):
        return self._extract_limited_data(self.time_candles, limit, max_limit=self.time_candles_index)

    def get_symbol_volume_candles(self, limit=-1):
        return self._extract_limited_data(self.volume_candles, limit, max_limit=self.volume_candles_index)

    def get_symbol_prices(self, limit=-1):
        return {
            enums.PriceIndexes.IND_PRICE_CLOSE.value: self.get_symbol_close_candles(limit),
            enums.PriceIndexes.IND_PRICE_OPEN.value: self.get_symbol_open_candles(limit),
            enums.PriceIndexes.IND_PRICE_HIGH.value: self.get_symbol_high_candles(limit),
            enums.PriceIndexes.IND_PRICE_LOW.value: self.get_symbol_low_candles(limit),
            enums.PriceIndexes.IND_PRICE_VOL.value: self.get_symbol_volume_candles(limit),
            enums.PriceIndexes.IND_PRICE_TIME.value: self.get_symbol_time_candles(limit)
        }

    def get_candles(self, limit=-1):
        candles_size = self.close_candles_index if limit == -1 else limit
        candles = [[]] * candles_size
        iter_range = range(self.close_candles_index) if limit == -1 \
            else range(self.close_candles_index-limit, self.close_candles_index)
        candles_index = 0
        for index in iter_range:
            candle = [0] * len(enums.PriceIndexes)
            candle[enums.PriceIndexes.IND_PRICE_CLOSE.value] = self.close_candles[index]
            candle[enums.PriceIndexes.IND_PRICE_OPEN.value] = self.open_candles[index]
            candle[enums.PriceIndexes.IND_PRICE_HIGH.value] = self.high_candles[index]
            candle[enums.PriceIndexes.IND_PRICE_LOW.value] = self.low_candles[index]
            candle[enums.PriceIndexes.IND_PRICE_VOL.value] = self.volume_candles[index]
            candle[enums.PriceIndexes.IND_PRICE_TIME.value] = self.time_candles[index]
            candles[candles_index] = candle
            candles_index += 1
        return candles

    def replace_all_candles(self, all_candles_data):
        self._reset_candles()
        self._set_all_candles(all_candles_data)
        self.candles_initialized = True

    def upsert_candle(self, updated_candle):
        updated_candle_time = updated_candle[enums.PriceIndexes.IND_PRICE_TIME.value]
        for index, candle_time in enumerate(self.time_candles):
            if candle_time == updated_candle_time:
                self.close_candles[index] = updated_candle[enums.PriceIndexes.IND_PRICE_CLOSE.value]
                self.open_candles[index] = updated_candle[enums.PriceIndexes.IND_PRICE_OPEN.value]
                self.high_candles[index] = updated_candle[enums.PriceIndexes.IND_PRICE_HIGH.value]
                self.low_candles[index] = updated_candle[enums.PriceIndexes.IND_PRICE_LOW.value]
                self.volume_candles[index] = updated_candle[enums.PriceIndexes.IND_PRICE_VOL.value]
                return

        # candle not in db, add it
        self.add_new_candle(updated_candle)

    def add_old_and_new_candles(self, candles_data):
        """
        Same as add_new_candle but also checks if old candles are missing
        :param candles_data: new candles data
        :return:
        """
        # check old candles
        for old_candle in candles_data[:-1]:
            if old_candle[enums.PriceIndexes.IND_PRICE_TIME.value] not in self.time_candles:
                self.add_new_candle(old_candle)

        try:
            self.add_new_candle(candles_data[-1])
        except IndexError as e:
            self.logger.error(f"Fail to add last candle {candles_data} : {e}")

    def add_new_candle(self, new_candle_data):
        """
        :param new_candle_data: new candles data
        :return:
        """
        if self._should_add_new_candle(new_candle_data[enums.PriceIndexes.IND_PRICE_TIME.value]):
            try:
                self._check_max_candles()
                self.close_candles[self.close_candles_index] = new_candle_data[enums.PriceIndexes.IND_PRICE_CLOSE.value]
                self.open_candles[self.open_candles_index] = new_candle_data[enums.PriceIndexes.IND_PRICE_OPEN.value]
                self.high_candles[self.high_candles_index] = new_candle_data[enums.PriceIndexes.IND_PRICE_HIGH.value]
                self.low_candles[self.low_candles_index] = new_candle_data[enums.PriceIndexes.IND_PRICE_LOW.value]
                self.time_candles[self.time_candles_index] = float(new_candle_data[enums.PriceIndexes.IND_PRICE_TIME.value])
                self.volume_candles[self.volume_candles_index] = new_candle_data[enums.PriceIndexes.IND_PRICE_VOL.value]
                self._inc_candle_index()
            except IndexError as e:
                self.logger.error(f"Fail to add new candle {new_candle_data} : {e}")

    # private
    def _set_all_candles(self, new_candles_data):
        if isinstance(new_candles_data[-1], list):
            for candle_data in new_candles_data:
                self.add_new_candle(candle_data)
        else:
            self.add_new_candle(new_candles_data)

    def _change_current_candle(self):
        self.close_candles = data_util.shift_value_array(self.close_candles, -1, np.nan, np.float64)
        self.open_candles = data_util.shift_value_array(self.open_candles, -1, np.nan, np.float64)
        self.high_candles = data_util.shift_value_array(self.high_candles, -1, np.nan, np.float64)
        self.low_candles = data_util.shift_value_array(self.low_candles, -1, np.nan, np.float64)
        self.volume_candles = data_util.shift_value_array(self.volume_candles, -1, np.nan, np.float64)
        self.time_candles = data_util.shift_value_array(self.time_candles, -1, np.nan, np.float64)

    def _should_add_new_candle(self, new_open_time):
        return new_open_time not in self.time_candles

    def _check_max_candles(self):
        if self.reached_max:
            self._change_current_candle()

    def _inc_candle_index(self):
        if self.close_candles_index < self.max_candles_count - 1:
            self.close_candles_index += 1
            self.open_candles_index += 1
            self.high_candles_index += 1
            self.low_candles_index += 1
            self.time_candles_index += 1
            self.volume_candles_index += 1
        else:
            self.reached_max = True

    def _extract_limited_data(self, data, limit=-1, max_limit=-1):
        max_handled_limit: int = self.max_candles_count if self.reached_max else max_limit
        if limit == -1:
            if max_limit == -1:
                return np.array(data, dtype=np.float64)
            return np.array(data[:max_handled_limit], dtype=np.float64)

        if max_limit == -1:
            return np.array(data[-min(limit, len(data)):], dtype=np.float64)
        else:
            return np.array(data[max(0, max_handled_limit - limit): max_handled_limit], dtype=np.float64)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_trading.exchanges import adapters
from octobot_trading.exchanges.adapters import (
    AbstractAdapter,
)
from octobot_trading.exchanges import exchanges
from octobot_trading.exchanges.exchanges import (
    ExchangeConfiguration,
    Exchanges,
)
from octobot_trading.exchanges import exchange_channels
from octobot_trading.exchanges.exchange_channels import (
    requires_refresh_trigger,
    create_exchange_channels,
    create_exchange_producers,
    create_authenticated_producer_from_parent,
)

from octobot_trading.exchanges import abstract_exchange
from octobot_trading.exchanges.abstract_exchange import (
    AbstractExchange,
)

from octobot_trading.exchanges import abstract_websocket_exchange
from octobot_trading.exchanges.abstract_websocket_exchange import (
    AbstractWebsocketExchange,
)

from octobot_trading.exchanges import exchange_manager
from octobot_trading.exchanges.exchange_manager import (
    ExchangeManager,
)

from octobot_trading.exchanges import exchange_factory
from octobot_trading.exchanges.exchange_factory import (
    create_exchanges,
    create_real_exchange,
    initialize_real_exchange,
    create_simulated_exchange,
    init_simulated_exchange,
)
from octobot_trading.exchanges.util import (
    ExchangeMarketStatusFixer,
    is_ms_valid,
    SymbolDetails,
    get_rest_exchange_class,
    get_order_side,
    log_time_sync_error,
    get_partners_explanation_message,
    get_enabled_exchanges,
    is_compatible_account,
    get_historical_ohlcv,
    get_exchange_type,
    get_default_exchange_type,
    get_supported_exchange_types,
    update_raw_order_from_raw_trade,
    is_missing_trading_fees,
    apply_trades_fees,
    get_common_traded_quote,
    get_associated_symbol,
    get_exchange_class_from_name,
    get_local_exchange_manager,
    get_auto_filled_exchange_names,
    get_exchange_details,
    force_disable_web_socket,
    check_web_socket_config,
    search_websocket_class,
    supports_websocket,
)
from octobot_trading.exchanges import exchange_websocket_factory
from octobot_trading.exchanges.exchange_websocket_factory import (
    is_channel_managed_by_websocket,
    is_channel_fully_managed_by_websocket,
    is_websocket_feed_requiring_init,
    search_and_create_websocket,
)
from octobot_trading.exchanges import config
from octobot_trading.exchanges.config import (
    ExchangeConfig,
    BacktestingExchangeConfig,
)
from octobot_trading.exchanges import traders
from octobot_trading.exchanges.traders import (
    Trader,
    TraderSimulator,
)
from octobot_trading.exchanges import types
from octobot_trading.exchanges.types import (
    WebSocketExchange,
    RestExchange,
)
from octobot_trading.exchanges import implementations
from octobot_trading.exchanges.implementations import (
    DefaultWebSocketExchange,
    ExchangeSimulator,
    DefaultRestExchange,
)
from octobot_trading.exchanges import exchange_builder
from octobot_trading.exchanges.exchange_builder import (
    ExchangeBuilder,
    create_exchange_builder_instance,
)
from octobot_trading.exchanges import connectors
from octobot_trading.exchanges.connectors import (
    CCXTWebsocketConnector,
    CCXTConnector,
    CCXTAdapter,
    ExchangeSimulatorConnector,
    ExchangeSimulatorAdapter,
)
from octobot_trading.exchanges import exchange_details
from octobot_trading.exchanges.exchange_details import (
    ExchangeDetails,
)

__all__ = [
    "AbstractAdapter",
    "ExchangeConfig",
    "BacktestingExchangeConfig",
    "ExchangeManager",
    "ExchangeBuilder",
    "create_exchange_builder_instance",
    "create_exchanges",
    "create_real_exchange",
    "initialize_real_exchange",
    "create_simulated_exchange",
    "init_simulated_exchange",
    "ExchangeConfiguration",
    "Exchanges",
    "get_rest_exchange_class",
    "get_order_side",
    "log_time_sync_error",
    "get_partners_explanation_message",
    "get_enabled_exchanges",
    "is_compatible_account",
    "get_historical_ohlcv",
    "get_exchange_type",
    "get_default_exchange_type",
    "get_supported_exchange_types",
    "update_raw_order_from_raw_trade",
    "is_missing_trading_fees",
    "apply_trades_fees",
    "get_common_traded_quote",
    "get_associated_symbol",
    "get_exchange_class_from_name",
    "get_local_exchange_manager",
    "get_auto_filled_exchange_names",
    "get_exchange_details",
    "AbstractExchange",
    "is_channel_managed_by_websocket",
    "is_channel_fully_managed_by_websocket",
    "is_websocket_feed_requiring_init",
    "search_and_create_websocket",
    "requires_refresh_trigger",
    "create_exchange_channels",
    "create_exchange_producers",
    "create_authenticated_producer_from_parent",
    "TraderSimulator",
    "Trader",
    "DefaultWebSocketExchange",
    "DefaultRestExchange",
    "ExchangeSimulator",
    "CCXTWebsocketConnector",
    "WebSocketExchange",
    "RestExchange",
    "ExchangeMarketStatusFixer",
    "is_ms_valid",
    "SymbolDetails",
    "AbstractWebsocketExchange",
    "force_disable_web_socket",
    "check_web_socket_config",
    "search_websocket_class",
    "supports_websocket",
    "CCXTConnector",
    "CCXTAdapter",
    "ExchangeSimulatorConnector",
    "ExchangeSimulatorAdapter",
    "ExchangeDetails",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.asset as asset_class


class MarginAsset(asset_class.Asset):
    def __init__(self, name, available, total,
                 borrowed=constants.ZERO,
                 interest=constants.ZERO,
                 locked=constants.ZERO):
        super().__init__(name, available, total)
        self.borrowed = borrowed
        self.interest = interest
        self.locked = locked

    def __str__(self):
        return super().__str__() + " | " \
                                   f"Borrowed: {float(self.borrowed)} | " \
                                   f"Interest: {float(self.interest)} | " \
                                   f"Locked: {float(self.locked)}"

    def __eq__(self, other):
        if isinstance(other, MarginAsset):
            return self.available == other.available and self.total == other.total and \
                   self.borrowed == other.borrowed and self.interest == other.interest and self.locked == other.locked
        return False

    def _specific_restore_unavailable_from_other(self, other_asset):
        if other_asset.borrowed != constants.ZERO:
            self.borrowed = self.borrowed + other_asset.borrowed
        if other_asset.interest != constants.ZERO:
            self.interest = self.interest + other_asset.interest
        if other_asset.locked != constants.ZERO:
            self.locked = self.locked + other_asset.locked

    def update(self, available=constants.ZERO, total=constants.ZERO, borrowed=constants.ZERO,
               interest=constants.ZERO, locked=constants.ZERO):
        """
        Update asset portfolio
        :param available: the available delta
        :param total: the total delta
        :param borrowed: the borrowed delta
        :param interest: the interest delta
        :param locked: the locked delta
        :return: True if updated
        """
        if available == constants.ZERO and total == constants.ZERO and borrowed == constants.ZERO and \
                interest == constants.ZERO and locked == constants.ZERO:
            return False
        self.available += self._ensure_update_validity(self.available, available)
        self.total += self._ensure_update_validity(self.total, total)
        self.borrowed += self._ensure_update_validity(self.borrowed, borrowed)
        self.interest += self._ensure_update_validity(self.interest, interest)
        self.locked += self._ensure_update_validity(self.locked, locked)
        return True

    def set(self, available=constants.ZERO, total=constants.ZERO, borrowed=constants.ZERO,
            interest=constants.ZERO, locked=constants.ZERO):
        """
        Set available, total, initial_margin, margin_balance, position_initial_margin and maintenance_margin
        values for portfolio asset
        :param available: the available value
        :param total: the total value
        :param borrowed: the borrowed value
        :param interest: the interest value
        :param locked: the locked value
        :return: True if updated
        """
        if available == self.available and total == self.total and borrowed == self.borrowed and \
                interest == self.interest and locked == self.locked:
            return False
        self.available = available
        self.total = total
        self.borrowed = borrowed
        self.interest = interest
        self.locked = locked
        return True

    def reset(self):
        """
        Reset asset portfolio to zero
        """
        self.set(available=constants.ZERO, total=constants.ZERO, borrowed=constants.ZERO,
                 interest=constants.ZERO, locked=constants.ZERO)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.api as api

import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.errors as errors

import octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater as ohlcv_updater
import octobot_trading.util as util


class OHLCVUpdaterSimulator(ohlcv_updater.OHLCVUpdater):
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer
        self.exchange_name = self.channel.exchange_manager.exchange_name

        self.initial_timestamp = api.get_backtesting_current_time(self.channel.exchange_manager.exchange.backtesting)
        self.last_timestamp_pushed = 0
        self.time_consumer = None

        self.future_candle_time_frame = self.channel.exchange_manager.exchange_config.get_shortest_time_frame()
        self.future_candle_sec_length = enums.TimeFramesMinutes[self.future_candle_time_frame] * \
                                        constants.MINUTE_TO_SECONDS

        self.last_candles_by_pair_by_time_frame = {}
        self.require_last_init_candles_pairs_push = False
        self.traded_pairs = self._get_traded_pairs()
        self.traded_time_frame = self._get_time_frames()

    async def start(self):
        if not self.is_initialized:
            await self._initialize(False)
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            pushed_data = False
            for pair in self.traded_pairs:
                for time_frame in self.traded_time_frame:
                    # Use last_timestamp_pushed + 1 for inferior timestamp to avoid select of an already selected candle
                    # (selection is <= and >=)
                    # Use timestamp + self.future_candle_sec_length to include the future candle on the future candles
                    # time frame that will be sorted in exchange simulator for later uses.
                    ohlcv_data: list = await self.exchange_data_importer.get_ohlcv_from_timestamps(
                        exchange_name=self.exchange_name,
                        symbol=pair,
                        time_frame=time_frame,
                        inferior_timestamp=self.last_timestamp_pushed + 1,
                        superior_timestamp=timestamp + (self.future_candle_sec_length
                                                        if self.future_candle_time_frame is time_frame else 0)
                    )
                    if ohlcv_data:
                        pushed_data = await self._handle_ohlcv_data(ohlcv_data, time_frame, pair, timestamp)
                    elif self.require_last_init_candles_pairs_push:
                        # triggered on first iteration to initialize large candles that might be pushed much later
                        # otherwise but are required to complete TA evaluation
                        if time_frame.value in self.last_candles_by_pair_by_time_frame[pair]:
                            await self.push(time_frame,
                                            pair,
                                            [self.last_candles_by_pair_by_time_frame[pair][time_frame.value][-1]],
                                            partial=True)
                            pushed_data = True
            self.channel.exchange_manager.exchange.is_unreachable = not pushed_data

        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e}")
            await self.pause()
            await self.stop()
        except IndexError as e:
            self.logger.warning(f"Failed to access ohlcv_data : {e}")
        except Exception as e:
            self.logger.exception(e, True, f"Error when updating from timestamp: {e}")
        finally:
            self.last_timestamp_pushed = timestamp
            self.require_last_init_candles_pairs_push = False

    async def _handle_ohlcv_data(self, ohlcv_data, time_frame, pair, timestamp):
        has_future_candle = False
        if self.future_candle_time_frame is time_frame:
            if ohlcv_data[-1][-1][enums.PriceIndexes.IND_PRICE_TIME.value] == timestamp:
                # register future candle
                self.channel.exchange_manager.exchange.get_current_future_candles()[pair][time_frame.value] = \
                    ohlcv_data[-1][-1]
                # do not push future candle
                has_future_candle = True
            else:
                # if no future candle available
                # (end of backtesting of missing data: reset future candle)
                self.channel.exchange_manager.exchange.get_current_future_candles()[pair][time_frame.value] = None

            # There should always be at least 2 candles in read data, otherwise this means that
            # the exchange was down for some time. Consider it unreachable
            self.channel.exchange_manager.exchange.is_unreachable = len(ohlcv_data) < 2
        if not has_future_candle or len(ohlcv_data) > 1:
            # push current candle(s)
            candles = ohlcv_data[:-1] if has_future_candle else ohlcv_data
            await self.push(time_frame,
                            pair,
                            [ohlcv[-1] for ohlcv in candles],
                            partial=True)
            return True
        return False

    async def pause(self):
        await util.pause_time_consumer(self)

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        await util.resume_time_consumer(self, self.handle_timestamp)

    def _get_traded_pairs(self):
        return api.get_available_symbols(self.exchange_data_importer)

    def _get_time_frames(self):
        return self.channel.exchange_manager.exchange.get_time_frames(self.exchange_data_importer)

    async def _initialize_candles(self, time_frame, pair, should_retry):
        # fetch history
        ohlcv_data = None
        try:
            # only load candles starting from the star time of the backtesting
            ohlcv_data: list = await self.exchange_data_importer.get_ohlcv_from_timestamps(
                exchange_name=self.exchange_name,
                symbol=pair,
                time_frame=time_frame,
                limit=self.OHLCV_OLD_LIMIT,
                inferior_timestamp=self.initial_timestamp,
                superior_timestamp=self.initial_timestamp)
            candles_len = len(ohlcv_data)
            self.logger.info(f"Loaded pre-backtesting starting timestamp historical "
                             f"candles for: {pair} in {time_frame}: {candles_len} "
                             f"candle{'s' if candles_len > 1 else ''}")
        except Exception as e:
            self.logger.exception(e, True, f"Error while fetching historical candles: {e}")
        if pair not in self.last_candles_by_pair_by_time_frame:
            self.last_candles_by_pair_by_time_frame[pair] = {}
        if ohlcv_data:
            # init historical candles
            await self.channel.exchange_manager.get_symbol_data(pair) \
                .handle_candles_update(time_frame,
                                       [ohlcv[-1] for ohlcv in ohlcv_data],
                                       replace_all=True,
                                       partial=False,
                                       upsert=False)
            self.last_candles_by_pair_by_time_frame[pair][time_frame.value] = ohlcv_data[-1]
            self.require_last_init_candles_pairs_push = True
        # self.initial_timestamp - 1 to re-select this candle and push it when init step will be over
        self.last_timestamp_pushed = self.initial_timestamp - 1



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import async_channel.constants as channel_constants

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.enums as enums


class MarkPriceProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, mark_price, mark_price_source=enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value):
        await self.perform(symbol, mark_price, mark_price_source=mark_price_source)

    async def perform(self, symbol, mark_price, mark_price_source=enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value):
        try:
            if self.channel.get_filtered_consumers(symbol=channel_constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                if self.channel.exchange_manager.get_symbol_data(symbol).handle_mark_price_update(mark_price,
                                                                                                  mark_price_source):
                    # only send mark price if price got updated
                    # mark_price attribute access required to send calculation result
                    await self.send(cryptocurrency=self.channel.exchange_manager.exchange.get_pair_cryptocurrency(
                                        symbol),
                                    symbol=symbol,
                                    mark_price=self.channel.exchange_manager.get_symbol_data(
                                        symbol).prices_manager.mark_price)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, mark_price):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "mark_price": mark_price
            })


class MarkPriceChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = MarkPriceProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer
